Using GObject
=============



How Do I Learn GObject ?
-----------------------

Been there, done that. Here are my advices.

First, you need to bookmark the official documentation. It's the best place to
start, but also a good place to go back afterward. Things that don't make sense
at the first reading will become clear as you go further.

- [GObject Reference Manual](https://developer.gnome.org/gobject/stable/)

Second, there's one essential article that nicely sums up the GObject
construction process, and raise some important points around it. This article
you will have to read again and again.

- [A gentle introduction to gobject construction](https://blogs.gnome.org/desrt/2012/02/26/a-gentle-introduction-to-gobject-construction/)

At last, you need code example, although the code you find is not always
up-to-date with latest/best way to do things. Anyway. The main GObject-based
libraries are probably the best places to go to learn GObject. Grab the source
code, sharpen your grep, and there you go.

- [GTK+](https://developer.gnome.org/gtk3/stable/)
- [GStreamer](https://gstreamer.freedesktop.org/data/doc/gstreamer/head/gstreamer/html/)



Good Practices
--------------

### Follow The Conventions

The first place to learn about good practices is the GObject conventions page.
It's more than good practice actually. GObject expects you to follow a few
conventions, so that code generation tools (`glib-mkenum`) and other stuff can
actually work. Don't try to avoid that.

- [GObject Conventions](https://developer.gnome.org/gobject/stable/gtype-conventions.html)

### Use `G_DECLARE_*` And `G_DEFINE_*`

Do not clutter the header file with too much boilerplate, as it's often seen in
some old GObject-based code.

- [G_DECLARE_FINAL_DERIVABLE_TYPE](https://blogs.gnome.org/desrt/2015/01/27/g_declare_finalderivable_type/)

### Do The Minimum In `_init()`

There should be almost nothing in the `init()` function. Most of the time,
it just boils down to setting up the private pointer, if any.

For all the rest, it's better to implement it in the `constructed()` method.

- [A gentle introduction to gobject construction](https://blogs.gnome.org/desrt/2012/02/26/a-gentle-introduction-to-gobject-construction/)

### Always Use A Private Pointer

You will notice that in here, we never put anything in the object structure.
We ALWAYS use a private structure.

Using a private structure is particularly useful for derivable objects, since
for these objects, the structure MUST BE defined in the header (public) to allow
inheritance. But when the object is not derivable, and therefore the structure
is defined in the code (private), what the point of having a private struct ?
Answer: consistency, and less hassle when the objects change from final to
derivable, or the other way around.

- [Changing quickly between a final and derivable GObject class](https://blogs.gnome.org/swilmet/2015/10/10/changing-quickly-between-a-final-and-derivable-gobject-class/)

Also, we ALWAYS use a pointer named `priv` to access it. It's just a convention,
but it's a very common one, and doing it everywhere makes the whole code
consistent.

### Best Way To Set Properties

The correct way to set an object pointer. It's not as trivial as it seems.

- [Use g_set_object() to simplify (and safetyify) GObject property setters](https://tecnocode.co.uk/2014/12/19/use-g_set_object-to-simplify-and-safetyify-gobject-property-setters/)



Quick FAQ
---------

### How To Change A Construct-Only Property Via Inheritance

This questions is asked on StackOverflow:

<http://stackoverflow.com/q/16557905/776208>

Unfortunately, the answer given by Emmanuele Bassi is wrong, and it seems that
the only other solution is to hack around, as did the guy who asked the question.

There's another solution though: don't use construct-only properties if they are
to be modified by the inherited objects. Make them construct.

### How To Implement A Singleton

This is discussed here and there on the Net.

- [GObject API Reference](https://developer.gnome.org/gobject/unstable/gobject-The-Base-Object-Type.html#GObjectClass)
- [How to make a GObject singleton](https://blogs.gnome.org/xclaesse/2010/02/11/how-to-make-a-gobject-singleton/)

### How To Have Protected Fields

GObject doesn't provide any solution to have protected fields. I managed to do
without up to now, but just in case, here's a workaround. Basically, it's just
about defining the private structure in the header (therefore making it public),
and naming it 'protected' instead of 'private'.

- [GObject and protected fields – simple hack](http://codica.pl/2008/12/21/gobject-and-protected-fields-simple-hack/)



More Links
----------

Tips from a C++ programmer, I didn't read it actually, but it seemed interesting.

- [Few tips about GObject for C++ programmers ](http://sgros.blogspot.com/2016/01/few-tips-about-gobject-for-oo.html)



Been there, done that
---------------------

### Should I Set A Property Asynchronously ?

Sometimes, setting a property is just about setting a variable.
But at other times, it might trigger quite a bit of code execution.

For example, I have an 'enable' property, and when I set it to true,
it really enables the whole object, and possibly a lot of things happens.

For some technical reasons, I had to actually delay the execution of the
enable/disable code, with `g_idle_add()`. And, as you can expect, I changed
the value of my property only *after* this asynchronous code was executed.
Only at that moment the 'notify' signal was sent.

This seems OK at a first glance. But in practice, there's a problem.
The code that sets the property may not want to receive the notify signal,
and therefore use `g_signal_handler_block()` before setting the property,
and `g_signal_handler_unblock()` after setting it.

If I set the property asynchronously, and therefore send the notify signal
afterward, the code that set the property will receive the notify signal,
even if it tried to block it. So it creates a bit of an unexpected behavior
for the code setting the property.

So... Should I set a property asynchronously ? The answer is NO.

For this particular problem of the 'enable' property, I solved it by having
two properties:
- 'enabled', readwrite, which reflects how the object is _configured_.
   Therefore this property is set synchronously.
- 'state', readable only, which gives the current state of the object,
   and go from disabled to enabling to enabled. This property is tied to
   'enabled', but its value changes aynchronously.

With two properties, one can connect to the 'enabled' property to know how the
object is configured, and one can connect to the 'state' property to know the
real object state.

### Should I Define A Base Class For My Object Hierarchy ?

For example, GStreamer has its own `GstObject`. And at some point, I thought it
was cool, and furthermore it seemed that it would solve some of the problems
I faced. So I defined my own `GvObject` for a while. I then removed it quickly
enough. Here's my feedback from that experience.

First questions that arises: should all my objects inherit from GvObject ?
Or only those who have a solid reason for that ?

At first, I went for the latter, because I thought that if an object has no
reason to inherit GvObject, then it should remain as lightweight and simple as
possible, and inherit GObject.

So I ended up with a mix of objects that were either GObject derived, either
GvObject derived. It's a bit of a mess suddenly, because the separation
between both, and the reasons for choosing the right parent, was all very
unclear.

Then, serialization kicked in, and I had another dilemna. If the serialization
API targets GvObject, then GObject derived objects were left out, and couldn't
be serialized. On the other hand, if the API targeted GObject, I couldn't take
advantage of some of the specific features GvObject, that were actually very
helpful for serializing.

Which made me change my mind: if I define a GvObject base-class, then I should
embrace it completely, and have every object inherit it. It looked nice and
consistent now, but I lost a bit of the initial lightness in the process, and
I was wondering if it was worth it.

But still, I was happy because now the serialization was easier to implement,
since I embedded some convenient stuff for it within my base class GvObject.

But then, another problem kicked in. The UI part by implemented by inheriting
GTK widgets. How will I serialize this GObject derived objects, now that the
serialization API targets the GvObject type ? My only solution here was to
rewrite the UI, use composition instead of inheritance, so that my UI objects
could be derived from GvObject and be serializable.

I didn't do that though. Instead, I came back to GObject for the base class.
For all the problems that I solved with GvObject, I thought again and solved
them differently, and in a better way.

After all that, I draw some conclusions for myself:

- think twice before defining a base class for your object hierarchy. It has
  its good points, for sure, but also its drawbacks.
- if you work with GObject-based libraries, be aware that the base class you
  define can't help for objects from those libraries, or that you inherit from
  those libraries. Your lower denominator for the whole set of classes you deal
  with will always be GObject. SO, does you base type really helps, or does it
  get in the way ?
- ask yourself if you can't solve your problem with an interface instead.
  Interface are more flexible. Even if you use objects from other libraries,
  you can still inherit them, and implement an interface in the child.
